library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.dcs_interface_pkg.all;

entity apb_to_dcs is
generic(
	arm_addr_space : integer := 1			--Converts ARM address spacing (0x4) to address space of 0x1 (not using 4LSBs of APB address)
											--1= ARM address, all other integers directly links address from APB to DCS
);
port(
	--APB input control signals
	penable				: in		std_logic;						-- APB enable signal. Asserted high on second pulse
	psel				: in		std_logic;						-- APB slave select from master
	pwrite				: in		std_logic;						-- APB direction setting
	
	--APB input and addr
	paddr				: in		std_logic_vector(31 downto 0);	-- APB address
	pwdata				: in		std_logic_vector(31 downto 0);	-- APB write data
	
	--APB output signals
	prdata				: out	std_logic_vector(31 downto 0);	-- APB read data
	pready				: out	std_logic;						-- APB hold signal, for read/write more than 2 cycles
	pslverr				: out	std_logic;						-- APB slave error signal

    clk                 : in    std_logic;                     -- global clock (40 MHz)
    reset_n             : in    std_logic;                     -- asynch reset, negative polarity
    reset_from_siu      : in    std_logic;                     -- asynch reset from SIU, positive polarity
	
	--internal resets
	global_reset        : out   std_logic;                     -- global reset, positive polarity
    rcu_reset           : out   std_logic;                     -- reset for RCU fw,  positive polarity
    fec_reset           : out   std_logic;                     -- reset for FEC, positive polarity
	
	--DCS bus signals
	we_dcs              : out   std_logic;                     -- write enable to the internal modules
	dcs_busBadd         : out   std_logic_vector(15 downto 0); -- address passed through
	dcs_busBdata_in     : in    std_logic_vector(31 downto 0); -- data from internal modules to DCS
	dcs_busBdata_out    : out   std_logic_vector(31 downto 0); -- data to internal modules from DCS
	dcs_bi              : out   std_logic;                     -- interrupt in case DCS needs bus
	dcs_bg              : in    std_logic;                     -- dcs bus grant from arbiter
    dcs_br              : out   std_logic;	                   -- dcs bus request to arbiter
    siu_bg              : in    std_logic                      -- siu bus grant from arbiter
	);
end apb_to_dcs;

architecture arc of apb_to_dcs is

--signal declarations
	type   state is (s_idle, s_wait_for_grant, s_grant, s_error);
	signal current_state, next_state : state;
	signal timeout			:std_logic;
	signal timeout_cnt		:std_logic_vector(6 downto 0);
	signal timeout_cnt_en	:std_logic;
	signal we_dcs_i			:std_logic;
	
begin

--The following is controlled by generic setting
arm_linking: if (arm_addr_space = 1) generate
--Linking to adapt to microcontroller addressing space
	dcs_busBadd <= paddr(19 downto 4);
end generate arm_linking;

no_arm_linking: if not(arm_addr_space = 1) generate
--Standard address linking
	dcs_busBadd <= paddr(15 downto 0);
end generate no_arm_linking;

timeout		<= timeout_cnt(6);
we_dcs				<= we_dcs_i;
dcs_br				<= '1' when (psel = '1'and (next_state = s_wait_for_grant or next_state = s_grant)) else '0';
dcs_busBdata_out	<= pwdata when (pwrite='1' and dcs_bg= '1') else (others => 'Z');
prdata				<= dcs_busBdata_in when (pwrite ='0' and dcs_bg= '1') else (others => 'Z');	

-- purpose: timeout counter for transaction
p_timeout_cnt: process(clk, reset_n, reset_from_siu)
begin
  if (reset_n = '0' or reset_from_siu = '1') then
    timeout_cnt <= (others => '0');
  elsif rising_edge(clk) then
    if (timeout_cnt_en = '1') then
	    timeout_cnt <= timeout_cnt + 1;
	  else
	    timeout_cnt <= (others => '0');
	  end if;
  end if;
end process p_timeout_cnt;

-- purpose: state machine driver
p_state_driver: process(clk, reset_n, reset_from_siu)
begin
  if (reset_n = '0' or reset_from_siu = '1') then
    current_state <= s_idle;
  elsif rising_edge(clk) then
    current_state <= next_state;
  end if;
end process p_state_driver;	

--purpose: set next state	
p_next_state: process(current_state, dcs_bg, timeout, psel, penable)		
begin
  case current_state is
    when s_idle =>
	  if (dcs_bg = '1' and psel = '1' and penable = '0') then		-- giving two cycle read/write possibility
		next_state <= s_grant;
      elsif (dcs_bg = '0' and psel = '1' and penable = '0') then
	    next_state <= s_wait_for_grant;
	  else
	    next_state <= s_idle;
	  end if;
  
    when s_wait_for_grant => 
      if (dcs_bg = '1') then 
  	    next_state <= s_grant;
	    elsif (timeout = '1') then
	      next_state <= s_error;
	    else
	      next_state <= s_wait_for_grant;
      end if;
  
    when  s_grant =>
      next_state <= s_idle;
   
    when s_error =>
      next_state <= s_idle;
    
	when others =>
	  next_state <= s_idle;
  end case;
end process p_next_state;	
	
-- purpose: set outputs of module and internal signals
p_output: process(clk, reset_n, reset_from_siu)
begin
  if (reset_n = '0' or reset_from_siu = '1') then
    timeout_cnt_en  <= '0';
	we_dcs_i        <= '0';     
	pslverr			<= '0';
	pready			<= '1';

  elsif rising_edge(clk) then
    case current_state is
    	when s_idle => 
			pslverr <= '0';
			timeout_cnt_en  <= '0';    
    	when s_wait_for_grant =>
		when s_grant =>
      		we_dcs_i        <= '0';
	  		pready			<= '1'; 	  
		when s_error => 
			pslverr <= '1';
			pready	<= '1';
			timeout_cnt_en <= '0';
		when others =>
	    --defaults
    end case;

  	case next_state is
		when s_idle =>
			pready <= '1';
		when s_wait_for_grant =>
			timeout_cnt_en  <= '1';
			pready <= '0';
		when s_grant =>
    		timeout_cnt_en  <= '0';
			pready <= '1';
			we_dcs_i <= pwrite;
		when others =>
	end case;

  end if;
end process p_output;	

--purpose: resets
p_reset : process(clk, reset_n)
begin
  if (reset_n = '0') then
    global_reset    <= '1'; -- send reset when the circuit is being globally reseted by the reset line
    rcu_reset       <= '1';  
    fec_reset       <= '1'; 
  elsif rising_edge(clk) then
    global_reset    <= '0'; -- send reset when the circuit is being globally reseted by the reset line
    rcu_reset       <= '0';  
    fec_reset       <= '0'; 
    if (we_dcs_i = '1') then
      case (paddr(15 downto 0)) is
        when c_global_reset =>
          global_reset <= '1';
        when c_rcu_reset =>
          rcu_reset    <= '1';
        when c_fec_reset =>
          fec_reset    <= '1';
        when others =>
          -- do nothing
      end case;
    end if;      
  end if;
end process p_reset;

--Purpose: Set up bus interrupt
p_bus_int: process (clk, reset_n, reset_from_siu)
begin
	if( reset_n = '0' or reset_from_siu = '1') then
		dcs_bi <= '0';
	elsif (rising_edge(clk)) then
		if(siu_bg = '0' or dcs_bg = '1') then		--do not interrupt if we have grant
			dcs_bi <= '0';
		elsif (paddr (15 downto 0) = c_arbiter_irq) then
			dcs_bi <= '1';
		end if;
	end if;
end process p_bus_int;
	
end architecture arc;