-------------------------------------------------------------------------------
-- Title      : Counters
-- Project    : RCU Trigger Receiver
-------------------------------------------------------------------------------
-- File             : $RCSfile: counters.vhd,v $
-- Last edited by   : $Author: alme $
-- Last update      : $Date: 2009/10/19 05:13:29 $
-- Current Revision : $Revision: 1.14 $
-------------------------------------------------------------------------------
-- Description: Trigger and error counters for the serial B message reception.
--              Generates time windows for valid trigger period as well.
--              Stores bunchcounter value when L1 arrives. 
-------------------------------------------------------------------------------
-- Revision History : 
-- http://www.ift.uib.no/kjekscgi-bin/viewcvs.cgi/vhdlcvs/trigger_receiver/
-------------------------------------------------------------------------------
--
-- This file is property of and copyright by the Instrumentation and 
-- Electronics Section, Dep. of Physics and Technology
-- University of Bergen, Norway, 2005
-- This file has been written by Johan Alme
-- Johan.Alme@ift.uib.no
--
-- Permission to use, copy, modify and distribute this firmware and its  
-- documentation strictly for non-commercial purposes is hereby granted  
-- without fee, provided that the above copyright notice appears in all  
-- copies and that both the copyright notice and this permission notice  
-- appear in the supporting documentation. The authors make no claims    
-- about the suitability of this firmware for any purpose. It is         
-- provided "as is" without express or implied warranty.                 
--
-------------------------------------------------------------------------------


-- UNDER UTVIKLING
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library work;
use work.trigger_receiver_pkg.all;

entity counters is
  port(	
    clk                 : in  std_logic;
    reset_n             : in  std_logic;
    module_reset        : in  std_logic;
    -- triggers to be counter
    pre_pulse           : in  std_logic;
    L0                  : in  std_logic;
    L1                  : in  std_logic;
    L1_msg              : in  std_logic;
    L2A                 : in  std_logic;
    L2R                 : in  std_logic;
    init_sequence       : in  std_logic; -- high when ok arrival of L0 and L1 if L0 is missing.
    receiving_L1        : in  std_logic; -- used for setting tw_counter when receiving messages
	  receiving_L2        : in  std_logic; 
	  busy                : in  std_logic;
    counter_reset       : in  std_logic;
    reset_bunch         : in  std_logic;
    single_bit_error    : in  std_logic;
    double_bit_error    : in  std_logic;
    msg_error           : in  std_logic; -- used for counting message  errors
    sequence_error      : in  std_logic; -- used for counting sequence errors
    L1_Latency          : in  std_logic_vector(11 downto 0);
    L1_boundary         : in  std_logic_vector(3 downto 0);
    L2_Latency          : in  std_logic_vector(31 downto 0);
    L1_msg_Latency      : in  std_logic_vector(31 downto 0);
    L0_tw               : out std_logic;
    L1_tw               : out std_logic;
    L1_exact_tw         : out std_logic;
    L1_msg_tw           : out std_logic;
    L2_tw               : out std_logic;
    L1_tw_passed        : out std_logic;--high if time window has passed.
    L1_msg_tw_passed    : out std_logic;
    L2_tw_passed        : out std_logic;
    BunchCnt_overflow   : out std_logic;
    Bunch_Counter       : out std_logic_vector(11 downto 0);
    BCID_Local          : out std_logic_vector(11 downto 0);
    pp_Counter          : out std_logic_vector(15 downto 0);
    L0_Counter          : out std_logic_vector(15 downto 0);
    L1_Counter          : out std_logic_vector(15 downto 0);
    L1_msg_Counter      : out std_logic_vector(15 downto 0);
    L2a_Counter         : out std_logic_vector(15 downto 0);
    L2r_Counter         : out std_logic_vector(15 downto 0);
    hammingErrorCnt     : out std_logic_vector(31 downto 0);
    msgErrorCnt         : out std_logic_vector(15 downto 0);
    seqErrorCnt         : out std_logic_vector(15 downto 0)
  );
END counters;


architecture arc of counters is

  signal Bunch_Counter_i         : std_logic_vector(12 downto 0);
  signal sbError_Count           : std_logic_vector(15 downto 0);
  signal dbError_Count           : std_logic_vector(15 downto 0);
  signal msgErrorCnt_i           : std_logic_vector(15 downto 0);
  signal seqErrorCnt_i           : std_logic_vector(15 downto 0);
 
  
  signal pp_Counter_i            : std_logic_vector(15 downto 0);
  signal L0_Counter_i            : std_logic_vector(15 downto 0);
  signal L1_Counter_i            : std_logic_vector(15 downto 0);
  signal L1_msg_Counter_i        : std_logic_vector(15 downto 0);
  signal L2a_Counter_i           : std_logic_vector(15 downto 0);
  signal L2r_Counter_i           : std_logic_vector(15 downto 0);
    
  signal tw_counter              : std_logic_vector(15 downto 0);
  signal L0R                     : std_logic;
  signal L1R                     : std_logic;
  
  signal BCID_Local_i            : std_logic_vector(11 downto 0);
  
  signal L0_tw_i                 : std_logic;
  signal L1_tw_i                 : std_logic;
  signal L1_exact_tw_i           : std_logic;
  signal L1_msg_tw_i             : std_logic;
  signal L2_tw_i                 : std_logic;
  signal L1_tw_passed_i          : std_logic;
  signal L1_msg_tw_passed_i      : std_logic;
  signal L2_tw_passed_i          : std_logic;
    
begin


  -- assigning outputs
  Bunch_Counter     <= Bunch_Counter_i(11 downto 0);
  msgErrorCnt       <= msgErrorCnt_i;
  seqErrorCnt       <= seqErrorCnt_i;
  pp_Counter        <= pp_Counter_i;
  L0_Counter        <= L0_Counter_i;
  L1_Counter        <= L1_Counter_i;
  L1_msg_Counter    <= L1_msg_Counter_i;
  L2a_Counter       <= L2a_counter_i;
  L2r_Counter       <= L2r_Counter_i;
  
  -- Purpose: Counts the number of errors.
  p_error_counters : process(clk, reset_n)
  begin
    if (reset_n='0') then
      msgErrorCnt_i <= (others => '0');
      seqErrorCnt_i <= (others => '0');
      sbError_Count <= (others => '0'); 
      dbError_Count <= (others => '0'); 
    elsif (rising_edge(clk)) then
      if (counter_reset = '1' or module_reset = '1') THEN
        msgErrorCnt_i <= (others => '0');
        seqErrorCnt_i <= (others => '0');
        sbError_Count <= (others => '0');
        dbError_Count <= (others => '0'); 
      else
        msgErrorCnt_i <= msgErrorCnt_i + msg_error;
        seqErrorCnt_i <= seqErrorCnt_i + sequence_error;
        sbError_Count <= sbError_Count + single_bit_error;
        dbError_Count <= dbError_Count + double_bit_error;
      end if;
    end if;
  end process p_error_counters;
 
  hammingErrorCnt   <= dbError_Count & sbError_Count;

-- Purpose: Counts number of bunchcrossings.
--          Reseted by bunchcount reset, decoded from serial B 
--          broadcast message.
p_bunchcounter: process(clk, reset_n)
begin
  if (reset_n = '0') then
    Bunch_Counter_i   <= (others => '0');
    BunchCnt_overflow <= '0';
   elsif (rising_edge(clk)) then
    if (reset_bunch = '1' or module_reset = '1') then
      Bunch_Counter_i   <= (others => '0');
      BunchCnt_overflow <= '0';
    else
      Bunch_Counter_i <= Bunch_Counter_i + 1;
      if (Bunch_Counter_i(12) = '1') then
        BunchCnt_overflow <= '1';
      end if;
    end if;
  end if;
end process p_bunchcounter;

-- Purpose: Counts the number different triggers
p_triggercounter: process(clk, reset_n)
begin
  if (reset_n= '0') then
    pp_Counter_i     <= (others => '0');
    L0_Counter_i     <= (others => '0');
    L1_Counter_i     <= (others => '0');
    L1_msg_Counter_i <= (others => '0');
    L2a_Counter_i    <= (others => '0');      
    L2r_Counter_i    <= (others => '0');      
  elsif (rising_edge(clk)) then
    if (counter_reset = '1' or module_reset = '1') then
      pp_Counter_i     <= (others => '0');
      L0_Counter_i     <= (others => '0');
      L1_Counter_i     <= (others => '0');
      L1_msg_Counter_i <= (others => '0');
      L2a_Counter_i    <= (others => '0');      
      L2r_Counter_i    <= (others => '0');      
    else
      pp_Counter_i     <= pp_Counter_i + pre_pulse;
      L0_Counter_i     <= L0_Counter_i + L0;
      L1_Counter_i     <= L1_Counter_i + L1;
      L1_msg_Counter_i <= L1_msg_Counter_i + L1_msg;
      L2a_Counter_i    <= L2a_counter_i + L2a;      
      L2r_Counter_i    <= L2r_Counter_i + L2r;      
    end if;
  end if;
end process p_triggercounter;


 
-- purpose: set up the legal time windows for the triggers
-- Also to store the local BCID including offset
p_timewindow_counter: process (clk, reset_n)
begin
  if (reset_n = '0') then
    tw_counter          <= (others => '0');
    L0R                 <= '0';
    L1R                 <= '0';
    
  elsif rising_edge(clk) then
    -- delaying triggers to match init_sequence pulse
    L0R        <= L0;
    L1R        <= L1;
    -- counter. 
    if (init_sequence = '1') then -- high on valid starting point on sequence.
      -- the latencies are added + 2 to account for the delay
      if (L0R = '1') then -- resets counter to L0 latency 
        tw_counter <= X"0002";
      elsif (L1R = '1') then
        tw_counter <= (X"0" & L1_Latency) + X"0002";
      end if;
    elsif (busy = '0' and receiving_L1 = '1') then
      tw_counter <=	L1_msg_Latency(31 downto 16) + X"0002";
	  elsif (busy = '0' and receiving_L2 = '1') then
      tw_counter <=	L2_Latency(31 downto 16) + X"0002";
	  elsif (busy = '1') then
      tw_counter <= tw_counter + 1;
    else
      tw_counter <= (others => '0');
    end if; 
  end if;
end process p_timewindow_counter;   

p_BCID_local : process(clk, reset_n)
begin
  if (reset_n = '0') then
    BCID_Local_i <= (others => '0');
      
  elsif rising_edge(clk) then
    if (init_sequence = '1' or module_reset = '1') then -- sampling at start of sequence
      BCID_Local_i <= (others => '0');
      if (L1R = '1') then
        BCID_Local_i <= Bunch_Counter_i(11 downto 0) - 1;
      end if;
    elsif (busy = '1') then -- sampling during sequence
      if (L1 = '1' and L1_tw_i = '1') then
        BCID_Local_i <= Bunch_Counter_i(11 downto 0);
      end if;
    end if; 
  end if;
end process p_BCID_local;


              
p_timewindow: process(clk, reset_n)
begin
  if (reset_n = '0') then
    L0_tw_i            <= '0';
    L1_tw_i            <= '0';
    L1_tw_passed_i     <= '0';
    L1_exact_tw_i      <= '0';
    L1_msg_tw_i        <= '0';
    L1_msg_tw_passed_i <= '0';
    L2_tw_i            <= '0';
    L2_tw_passed_i     <= '0';
    
  elsif rising_edge(clk) then
    L1_exact_tw_i      <= '0';
    L0_tw_i            <= '0';
    L1_tw_i            <= '0';    
    -- set up time windows
    if (busy = '0') then --If no seqeunce are ongoing both L0 and L1 time windows are high( since both are aloud)
      L0_tw_i            <= '1';
      L1_tw_i            <= '1';
      L1_tw_passed_i     <= '0';
      L1_exact_tw_i      <= '0';
      L1_msg_tw_i        <= '0';
      L1_msg_tw_passed_i <= '0';
      L2_tw_i            <= '0';
      L2_tw_passed_i     <= '0';
    else
      --  L1 time window - exact setting and boundary conditions
      if (tw_counter = L1_Latency) then 
        L1_exact_tw_i <= '1';
      end if;
      -- L1 trigger time window
      if (tw_counter >= (L1_Latency - L1_boundary)) then
        L1_tw_i <= '1';
      end if;
      if (tw_counter = (L1_Latency + L1_boundary + 1)) then
        L1_tw_passed_i <= '1';
      end if; 
      -- L1 msg time window:
      if (tw_counter = L1_msg_Latency(31 downto 16)) then
        L1_msg_tw_i <= '1';
      end if;
      if (tw_counter = L1_msg_Latency(15 downto 0)) then
        L1_msg_tw_passed_i <= '1';
      end if;
      -- L2 message window
      if (tw_counter = L2_Latency(31 downto 16)) then
        L2_tw_i <= '1';
      end if;
      if (tw_counter = L2_Latency(15 downto 0)) then
        L2_tw_passed_i <= '1';
      end if;
    end if;
  end if;
end process p_timewindow;


L0_tw               <= L0_tw_i;
L1_tw               <= L1_tw_i and not L1_tw_passed_i;
L1_exact_tw         <= L1_exact_tw_i;
L1_msg_tw           <= L1_msg_tw_i and not L1_msg_tw_passed_i;
L2_tw               <= L2_tw_i and not L2_tw_passed_i;
L1_tw_passed        <= L1_tw_passed_i;
L1_msg_tw_passed    <= L1_msg_tw_passed_i;
L2_tw_passed        <= L2_tw_passed_i;
BCID_Local          <= BCID_Local_i;
 
end arc;