-------------------------------------------------------------------------------
-- Title      : FIFO Wrapper
-- Project    : RCU Trigger Receiver
-------------------------------------------------------------------------------
-- File             : $RCSfile: fifo_wrapper.vhd,v $
-- Last edited by   : $Author: alme $
-- Last update      : $Date: 2008/02/05 11:42:01 $
-- Current Revision : $Revision: 1.9 $
-------------------------------------------------------------------------------
-- Description: Contains logic for writing data into the FIFO as well as 
--              counters to keep track of data.
-------------------------------------------------------------------------------
-- Revision History : 
-- http://www.ift.uib.no/kjekscgi-bin/viewcvs.cgi/vhdlcvs/trigger_receiver/
-------------------------------------------------------------------------------
--
-- This file is property of and copyright by the Instrumentation and 
-- Electronics Section, Dep. of Physics and Technology
-- University of Bergen, Norway, 2005
-- This file has been written by Johan Alme
-- Johan.Alme@ift.uib.no
--
-- Permission to use, copy, modify and distribute this firmware and its
-- documentation strictly for non-commercial purposes is hereby granted
-- without fee, provided that the above copyright notice appears in all
-- copies and that both the copyright notice and this permission notice
-- appear in the supporting documentation. The authors make no claims
-- about the suitability of this firmware for any purpose. It is
-- provided "as is" without express or implied warranty.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;

library work;
use work.trigger_receiver_pkg.all;

entity fifo_wrapper is
  generic (
    platform       : integer := 1; -- 0 = Xilinx, 1 = microsemi
    include_parity : integer := 1);
  port(
    clk                 : in  std_logic;
    reset_n             : in  std_logic;
    module_reset        : in  std_logic;
    --external interface
    read_enable         : in  std_logic;
    data_out            : out std_logic_vector(32 downto 0);
    read_counter_out    : out std_logic_vector(3 downto 0);
    buffered_events     : out std_logic_vector(3 downto 0);
    --internal interface
    event_info_reg      : in  std_logic_vector(31 downto 0);
    event_error_reg     : in  std_logic_vector(31 downto 0);
    DAQ_header          : in  DAQ_header_array(0 to 6);
    L2_trigger          : in  std_logic -- L2a or L2r/L2_timeout
  );
end fifo_wrapper;

architecture behave of fifo_wrapper is
  signal reset          : std_logic;
  signal data_in        : std_logic_vector(32 downto 0);
  signal write_enable   : std_logic;
  signal fifo_empty     : std_logic;
  signal fifo_full      : std_logic;
  signal L2_triggerR    : std_logic;
  signal event_counter  : std_logic_vector(3 downto 0);
  signal read_counter   : std_logic_vector(3 downto 0);
  signal write_counter  : std_logic_vector(3 downto 0);
  signal writing        : std_logic;
  
  signal cnt_dn        : std_logic; --- for debug reason stop fifo counts down when cnt up same time

  type FIFO_array  is array(integer range <>) of std_logic_vector(32 downto 0);
  signal data : FIFO_array(0 to 8);
 
begin

	cnt_dn <= '1' when read_counter = "1001" else '0';

  -- setting fifo control and outputs
  reset           <= (not reset_n) or module_reset;
  write_enable    <= writing;
  data_in         <= data(to_integer(unsigned(write_counter(3 downto 0))));
  buffered_events <= event_counter;
  read_counter_out <= read_counter;
  -- include or exclude parity bit generation for FIFO
  parity: if (include_parity = 1) generate
  begin
    data(0) <= parity32(event_info_reg)  & event_info_reg;
    data(1) <= parity32(event_error_reg) & event_error_reg;
    data(2) <= parity32(DAQ_header(0))   & DAQ_header(0); 
    data(3) <= parity32(DAQ_header(1))   & DAQ_header(1); 
    data(4) <= parity32(DAQ_header(2))   & DAQ_header(2); 
    data(5) <= parity32(DAQ_header(3))   & DAQ_header(3); 
    data(6) <= parity32(DAQ_header(4))   & DAQ_header(4); 
    data(7) <= parity32(DAQ_header(5))   & DAQ_header(5); 
    data(8) <= parity32(DAQ_header(6))   & DAQ_header(6); 
  end generate parity;

  not_parity: if not(include_parity = 1) generate
  begin
    data(0) <= '0' & event_info_reg;
    data(1) <= '0' & event_error_reg;
    data(2) <= '0' & DAQ_header(0); 
    data(3) <= '0' & DAQ_header(1); 
    data(4) <= '0' & DAQ_header(2); 
    data(5) <= '0' & DAQ_header(3); 
    data(6) <= '0' & DAQ_header(4); 
    data(7) <= '0' & DAQ_header(5); 
    data(8) <= '0' & DAQ_header(6); 
  end generate not_parity;  
  
  -- delays trigger for event counter
  p_delay_trigger : process(clk, reset_n)
  begin
    if (reset_n = '0') then
      L2_triggerR <= '0';
    elsif rising_edge(clk) then
      L2_triggerR <= L2_trigger;
    end if;
  end process p_delay_trigger;

  -- keep track of input data and number of events stored
  p_counters : process(clk, reset_n)
  begin
    if (reset_n = '0') then
      write_counter  <= (others => '0');
      read_counter   <= (others => '0');
      event_counter  <= (others => '0');
      writing        <= '0';
    elsif rising_edge(clk) then
      --default values
      write_counter  <= (others => '0');

      -- increasing write counter
      if (L2_trigger = '1' and fifo_full = '0') then
        --write_counter <= write_counter + 1;
        writing       <= '1';
      elsif (writing = '1') then
        write_counter <= write_counter + 1;
        if (write_counter(3 downto 0) = "1000") then
          write_counter <= (others => '0');
          writing       <= '0';
        end if;
      end if;

      if (read_counter = "1001" or module_reset = '1') then
        read_counter <= (others => '0');
      elsif (read_enable = '1' and fifo_empty = '0') then
        read_counter <= read_counter + 1;
      end if;
----------------------------------------------------------------
      if (module_reset = '1') then
        event_counter <= (others => '0');
      elsif (L2_triggerR = '1' and cnt_dn = '0') then
        event_counter <= event_counter + 1;
      elsif (L2_triggerR = '0' and cnt_dn = '1') then
        event_counter <= event_counter -1;
	  elsif (L2_triggerR = '1' and cnt_dn = '1') then
        event_counter <= event_counter ;
      end if;

    end if;
  end process p_counters;

  --xilinx_version: if (platform = 0) generate
--  begin
 --   CDH_fifo : entity work.XIL_event_fifo
	--    port map(
	 --     clk          => clk,
	  --    din          => data_in,
	   --   rd_en        => read_enable,
--	      rst          => reset,
--	      wr_en        => write_enable,
--	      dout         => data_out,
--	      empty        => fifo_empty,
--	      full         => fifo_full);
--  end generate xilinx_version;

  microsemi_version: if (platform = 1) generate
  begin
    CDH_fifo : entity work.ms_event_fifo
	    port map(
        -- Inputs
        CLK   => clk,--: in  std_logic;
        DATA  => data_in,--: in  std_logic_vector(32 downto 0);
        RE    => read_enable,--: in  std_logic;
        RESET => reset,--: in  std_logic;
        WE    => write_enable,--: in  std_logic;
        -- Outputs
        EMPTY => fifo_empty,--: out std_logic;
        FULL  => fifo_full,--: out std_logic;
        Q     => data_out);--: out std_logic_vector(32 downto 0)
   end generate microsemi_version;

end architecture behave;

