-------------------------------------------------------------------------------
-- Title      : Hamming Decoder
-- Project    : RCU Trigger Receiver
-------------------------------------------------------------------------------
-- File             : $RCSfile: hamming_decoder.vhd,v $
-- Last edited by   : $Author: alme $
-- Last update      : $Date: 2007/12/10 15:22:49 $
-- Current Revision : $Revision: 1.9 $
-------------------------------------------------------------------------------
-- Description: Hamming decoder for the Serial B line.
-------------------------------------------------------------------------------
-- Revision History : 
-- http://www.ift.uib.no/kjekscgi-bin/viewcvs.cgi/vhdlcvs/trigger_receiver/
-------------------------------------------------------------------------------
--
-- This file is property of and copyright by the Instrumentation and 
-- Electronics Section, Dep. of Physics and Technology
-- University of Bergen, Norway, 2005
-- This file has been written by Johan Alme
-- Johan.Alme@ift.uib.no
--
-- Permission to use, copy, modify and distribute this firmware and its  
-- documentation strictly for non-commercial purposes is hereby granted  
-- without fee, provided that the above copyright notice appears in all  
-- copies and that both the copyright notice and this permission notice  
-- appear in the supporting documentation. The authors make no claims    
-- about the suitability of this firmware for any purpose. It is         
-- provided "as is" without express or implied warranty.                 
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.trigger_receiver_pkg.all;

entity hamming_decoder is
  port(
    clk                 : in  std_logic;
    reset_n             : in  std_logic;
    module_reset        : in  std_logic;
    received_data       : in  std_logic_vector(38 downto 0);
    brcdata_ready_in    : in  std_logic;
    adrdata_ready_in    : in  std_logic;
    init_sequence       : in  std_logic;
    single_bit_error    : out std_logic;
    double_bit_error    : out std_logic;
    error_reg           : out std_logic_vector(3 downto 0);
    brcdata_ready_out   : out std_logic;
    adrdata_ready_out   : out std_logic;
    brcdata_out         : out std_logic_vector(7  downto 0);
    adrdata_out         : out std_logic_vector(31 downto 0)
  );
end hamming_decoder;

architecture behave of hamming_decoder is

  signal adr_data             : std_logic_vector(31 downto 0);
  signal brcst_data           : std_logic_vector(7 downto 0);
  signal syndrom_adrdata      : std_logic_vector(6 downto 0);
  signal syndrom_brcstdata    : std_logic_vector(4 downto 0);
  signal adr_single_bit_error : std_logic;
  signal adr_double_bit_error : std_logic;
  signal brcst_single_bit_error : std_logic;
  signal brcst_double_bit_error : std_logic;
  signal hmg_adrdata : std_logic_vector(6 downto 0);
  signal hmg_brcdata : std_logic_vector(4 downto 0);

  alias adr_single_bit_error_reg    : std_logic is error_reg(0);
  alias adr_double_bit_error_reg    : std_logic is error_reg(1);
  alias brcst_single_bit_error_reg  : std_logic is error_reg(2);
  alias brcst_double_bit_error_reg  : std_logic is error_reg(3);

begin

-- setting broadcast data and adressed data word from decoded received word
brcst_data <= received_data(12 downto 5);
adr_data   <= received_data(38 downto 7);
hmg_adrdata<= received_data(6 downto 0);
hmg_brcdata<= received_data(4 downto 0);
brcdata_ready_out <= brcdata_ready_in;
adrdata_ready_out <= adrdata_ready_in;



-- Calculating syndroms

syndrom_adrdata   <= hamming32(adr_data, hmg_adrdata)  xor hmg_adrdata;
syndrom_brcstdata <= hamming8(brcst_data, hmg_brcdata) xor hmg_brcdata;
-- hamming decoding the broadcast messages.
p_hmg_broadcast_decoder: process(brcst_data, syndrom_brcstdata)
begin
  brcdata_out        <=  brcst_data;
  brcst_double_bit_error <= '0';
  brcst_single_bit_error <= '0';
  -- hamming correcting
  if (syndrom_brcstdata(4) = '0' and (syndrom_brcstdata(3 downto 0) /= "0000")) then
    brcst_double_bit_error <= '1';
    brcst_single_bit_error <= '0';				
  else
    --default values
    brcst_double_bit_error <= '0';
    brcst_single_bit_error <= '1';		
    case syndrom_brcstdata(3 downto 0) is
      when "0000" => brcst_single_bit_error <= '0';
      when "0011" => brcdata_out(0)      <= not brcst_data(0); 
      when "1101" => brcdata_out(1)      <= not brcst_data(1); 
	    when "0101" => brcdata_out(2)      <= not brcst_data(2);
      when "1001" => brcdata_out(3)      <= not brcst_data(3);
      when "1110" => brcdata_out(4)      <= not brcst_data(4);
      when "0110" => brcdata_out(5)      <= not brcst_data(5);
      when "1010" => brcdata_out(6)      <= not brcst_data(6);
      when "1100" => brcdata_out(7)      <= not brcst_data(7);
      when others => brcst_double_bit_error <= '1';
    end case;		                 	
	end if;
end process p_hmg_broadcast_decoder;

-- hamming decoder of the individual addressed messages
p_hmg_addressed_decoder: process(adr_data, syndrom_adrdata) 
begin
  adrdata_out        <= adr_data;
  adr_double_bit_error  <= '0';
  adr_single_bit_error  <= '0';
      
	--hamming correcting	    
	if (syndrom_adrdata(6)='0' and (syndrom_adrdata(5 downto 0) /= "000000")) then
    adr_double_bit_error  <= '1';
    adr_single_bit_error  <= '0';
  else
    --default values
    adr_double_bit_error  <= '0';
    adr_single_bit_error  <= '1';
    case syndrom_adrdata(5 downto 0) is
      when "000000" => adr_single_bit_error    <= '0';		
      when "011001" => adrdata_out(0)      <= not adr_data(0);		        		
      when "101001" => adrdata_out(1)      <= not adr_data(1);		        	
      when "001001" => adrdata_out(2)      <= not adr_data(2);
      when "110001" => adrdata_out(3)      <= not adr_data(3);		        	
      when "010001" => adrdata_out(4)      <= not adr_data(4);		        			        			        			        			        			        	
      when "100001" => adrdata_out(5)      <= not adr_data(5);		        	
      when "111110" => adrdata_out(6)      <= not adr_data(6);		     
      when "011110" => adrdata_out(7)      <= not adr_data(7);		        	
      when "101110" => adrdata_out(8)      <= not adr_data(8);		        	
      when "001110" => adrdata_out(9)      <= not adr_data(9);		        	
      when "110110" => adrdata_out(10)     <= not adr_data(10);
      when "010110" => adrdata_out(11)     <= not adr_data(11);		        	
      when "100110" => adrdata_out(12)     <= not adr_data(12);		        			        			        			        			        			        			        			        	
      when "000110" => adrdata_out(13)     <= not adr_data(13);		        	
      when "111010" => adrdata_out(14)     <= not adr_data(14);		        	
      when "011010" => adrdata_out(15)     <= not adr_data(15);		        	
      when "101010" => adrdata_out(16)     <= not adr_data(16);		        	
      when "001010" => adrdata_out(17)     <= not adr_data(17);		        	
      when "110010" => adrdata_out(18)     <= not adr_data(18);		        	
      when "010010" => adrdata_out(19)     <= not adr_data(19);		        	
      when "100010" => adrdata_out(20)     <= not adr_data(20);		        	
      when "111100" => adrdata_out(21)     <= not adr_data(21);		        	
      when "011100" => adrdata_out(22)     <= not adr_data(22);		        	
      when "101100" => adrdata_out(23)     <= not adr_data(23);		        	
      when "001100" => adrdata_out(24)     <= not adr_data(24);		        	
      when "110100" => adrdata_out(25)     <= not adr_data(25);		        	
      when "010100" => adrdata_out(26)     <= not adr_data(26);
      when "100100" => adrdata_out(27)     <= not adr_data(27);		        	
      when "111000" => adrdata_out(28)     <= not adr_data(28);		        	
      when "011000" => adrdata_out(29)     <= not adr_data(29);		        	
      when "101000" => adrdata_out(30)     <= not adr_data(30);
      when "110000" => adrdata_out(31)     <= not adr_data(31);		        	
      when others =>   adr_double_bit_error <= '1';
    end case;
  end if;	             
end process p_hmg_addressed_decoder;


p_set_error_regs: process(clk, reset_n)
begin
  if (reset_n = '0') then
    --addressed messages
    adr_single_bit_error_reg <= '0';
    adr_double_bit_error_reg <= '0';
    --broadcast messages
    brcst_single_bit_error_reg <= '0';
    brcst_double_bit_error_reg <= '0';
  elsif rising_edge(clk) then
    if (module_reset = '1' or init_sequence = '1') then
      --addressed messages
      adr_single_bit_error_reg <= '0';
      adr_double_bit_error_reg <= '0';
      --broadcast messages
      brcst_single_bit_error_reg <= '0';
      brcst_double_bit_error_reg <= '0';
    else
      -- single adressed error
      if (adrdata_ready_in = '1' and adr_single_bit_error = '1') then
        adr_single_bit_error_reg <= '1';
      end if;
      --double adressed error
      if (adrdata_ready_in = '1' and adr_double_bit_error = '1') then
        adr_double_bit_error_reg <= '1';
      end if;
      --single broadcast error
      if (brcdata_ready_in = '1' and brcst_single_bit_error = '1') then
        brcst_single_bit_error_reg <= '1';
      end if;
      --double broadcast error
      if (brcdata_ready_in = '1' and brcst_double_bit_error = '1') then
        brcst_double_bit_error_reg <= '1';
      end if;
    end if;
  end if;
end process p_set_error_regs;

p_set_error_flags: process(clk, reset_n)
begin
  if (reset_n = '0') then
    single_bit_error    <= '0';		
    double_bit_error    <= '0';
  elsif rising_edge(clk) then
    -- default values
    single_bit_error <= '0';
    double_bit_error <= '0';

    if (module_reset = '1' or init_sequence = '1') then
      single_bit_error     <= '0';
      double_bit_error     <= '0';
    else
	    if (adrdata_ready_in = '1') then
        single_bit_error     <= adr_single_bit_error;
        double_bit_error     <= adr_double_bit_error;
      end if;
      if (brcdata_ready_in = '1') then
        single_bit_error     <= brcst_single_bit_error;
        double_bit_error     <= brcst_double_bit_error;
      end if;
    end if;
  end if;  
end process p_set_error_flags;


end architecture behave;

