-------------------------------------------------------------------------------
-- Title      : RCU communication Module
-- Project    : RCU Trigger Receiver
-------------------------------------------------------------------------------
-- File             : $RCSfile: rcu_com_release.vhd,v $
-- Last edited by   : $Author: alme $
-- Last update      : $Date: 2009/10/19 05:12:49 $
-- Current Revision : $Revision: 1.12 $
-------------------------------------------------------------------------------
-- Description      : Communicates with the CERN standard DCS interface. 
-------------------------------------------------------------------------------
-- Revision History : 
-- http://www.ift.uib.no/kjekscgi-bin/viewcvs.cgi/vhdlcvs/trigger_receiver/
-------------------------------------------------------------------------------
--
-- This file is property of and copyright by the Instrumentation and 
-- Electronics Section, Dep. of Physics and Technology
-- University of Bergen, Norway, 2005
-- This file has been written by Johan Alme
-- Johan.Alme@ift.uib.no
--
-- Permission to use, copy, modify and distribute this firmware and its  
-- documentation strictly for non-commercial purposes is hereby granted  
-- without fee, provided that the above copyright notice appears in all  
-- copies and that both the copyright notice and this permission notice  
-- appear in the supporting documentation. The authors make no claims    
-- about the suitability of this firmware for any purpose. It is         
-- provided "as is" without express or implied warranty.                 
--
-------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.trigger_receiver_pkg.all;


entity rcu_com_release is
  port(
    clk                 : in  std_logic;

    reset_n             : in  std_logic;
    --rcu bus
    data_in             : in    std_logic_vector(31 downto 0);
    data_out            : out   std_logic_vector(31 downto 0);
    RCU_addr            : in    std_logic_vector(15 downto 0);
    we                  : in    std_logic;
    -- output to internal logic
    -- signals from control register
    input_enable        : out   std_logic;
    L2a_mask            : out   std_logic;
    L2r_mask            : out   std_logic;
    L2_timeout_mask     : out   std_logic;
    counter_reset       : out   std_logic; -- reset all counters and error bits
    module_reset        : out   std_logic; -- reset all counters and statemachines etc. 
    disable_error_mask  : out   std_logic;
    trigger_input_mask_enable     : out   std_logic;
    L1a_msg_mask        : out   std_logic; -- high if only L1 header is enough
    L0_support          : out   std_logic; --support for L0 - default enabled
    header_version      : in    std_logic_vector( 3 downto 0);
    L1_Latency          : out   std_logic_vector(11 downto 0);
    L1_boundary         : out   std_logic_vector( 3 downto 0);
    L2_Latency          : out   std_logic_vector(31 downto 0);
    L2_ExtendedLatency    : out   std_logic_vector(31 downto 0);
    L1_msg_Latency      : out   std_logic_vector(31 downto 0);
    -- Counter values
    bc_overflow         : in    std_logic;
    busy                : in    std_logic;
    run_active          : in    std_logic;
    BCID_Local          : in    std_logic_vector(11 downto 0);
    pp_counter          : in    std_logic_vector(15 downto 0);
    L0_Counter          : in    std_logic_vector(15 downto 0);
    L1_Counter          : in    std_logic_vector(15 downto 0);
    L2a_Counter         : in    std_logic_vector(15 downto 0);
    L2r_Counter         : in    std_logic_vector(15 downto 0);
    L1_msg_counter      : in    std_logic_vector(15 downto 0);
    hammingErrorCnt     : in    std_logic_vector(31 downto 0);
    msgErrorCnt         : in    std_logic_vector(15 downto 0);
    seqErrorCnt         : in    std_logic_vector(15 downto 0);
    -- trigger info words, in eventheader format.
    buffered_events     : in    std_logic_vector( 3 downto 0);
    event_info_reg      : in    std_logic_vector(17 downto 0);
    event_error_reg     : in    std_logic_vector(24 downto 0);
    DAQ_header          : in    DAQ_header_array(0 to 6));
end rcu_com_release;

architecture arc of rcu_com_release is

-- registers and signals
  signal L1a_msg_mask_i       : std_logic := '1';
  signal L2a_mask_i           : std_logic := '1';
  signal L2r_mask_i           : std_logic := '1';
  signal L2_timeout_mask_i    : std_logic := '1';
  signal input_enable_i       : std_logic := '1';
  signal disable_error_mask_i : std_logic := '0';
  signal trigger_input_mask_enable_i    : std_logic := '0';
  signal L0_support_i         : std_logic := '1';
  signal L1_Latency_reg       : std_logic_vector(11 downto 0) := df_L1_latency;
  signal L1_boundary_reg      : std_logic_vector(3 downto 0)  := df_L1_boundary;
  signal L2_Latency_reg       : std_logic_vector(31 downto 0) := df_L2msg_latency_min & df_L2msg_latency_max;
  signal L2_ExtendedLatency_reg : std_logic_vector(31 downto 0) := df_L2_ExtendedLatency_min & df_L2_ExtendedLatency_max;
  signal L1_msg_Latency_reg   : std_logic_vector(31 downto 0) := df_L1msg_latency_min & df_L1msg_latency_max;
                                                       
begin
  
-- set outputs
L1a_msg_mask        <= L1a_msg_mask_i;
L2a_mask            <= L2a_mask_i;
L2r_mask            <= L2r_mask_i;
L2_timeout_mask     <= L2_timeout_mask_i;
input_enable        <= input_enable_i;
disable_error_mask  <= disable_error_mask_i;
trigger_input_mask_enable     <= trigger_input_mask_enable_i;
L0_support          <= L0_support_i;
L1_Latency          <= L1_Latency_reg;
L1_boundary         <= L1_boundary_reg;
L2_Latency          <= L2_Latency_reg;
L2_ExtendedLatency    <= L2_ExtendedLatency_reg;
L1_msg_Latency      <= L1_msg_Latency_reg;  
  
-- purpose: communcation with the DCS interface.
p_rcu_communication : process(clk, reset_n)
begin
  if (reset_n = '0') then
    L2a_mask_i           <= '1';
    L2r_mask_i           <= '1';
    L2_timeout_mask_i    <= '1';
    input_enable_i       <= '1';
    disable_error_mask_i <= '0';
    trigger_input_mask_enable_i    <= '0';
    L0_support_i         <= '1';
    L1a_msg_mask_i       <= '1';
    module_reset         <= '0';
    counter_reset        <= '0';
    L1_Latency_reg       <= df_L1_latency;
    L1_boundary_reg      <= df_L1_boundary;
    L2_Latency_reg       <= df_L2msg_latency_min & df_L2msg_latency_max;
    L2_ExtendedLatency_reg <= df_L2_ExtendedLatency_min & df_L2_ExtendedLatency_max;
    L1_msg_Latency_reg   <= df_L1msg_latency_min & df_L1msg_latency_max;
    data_out             <= (others => '0');
  elsif (rising_edge(clk)) then
    -- default values
    module_reset         <= '0';
    counter_reset        <= '0';
    if (we = '0' and rcu_addr(15 downto 12) = c_Module_address) then -- reading
      case rcu_addr(11 downto 0) is
        when c_control          => data_out <= c_version & header_version & '0' & busy & run_active & bc_overflow &  -- 16 MSB
                                               "000" & trigger_input_mask_enable_i & L1a_msg_mask_i & L2_timeout_mask_i & L2r_mask_i & L2a_mask_i & X"0" & L0_support_i & '0' & disable_error_mask_i & input_enable_i;
        when c_L1_Latency       => data_out <= X"0000" & L1_boundary_reg & L1_Latency_reg;
        when c_L2_Latency       => data_out <= L2_Latency_reg;
        when c_L2_ExtendedLatency => data_out <= L2_ExtendedLatency_reg;
        when c_L1_msg_Latency   => data_out <= L1_msg_Latency_reg;
        when c_pp_counter       => data_out <= X"0000" & pp_counter;
        when c_BCID_local       => data_out <= X"00000" & BCID_Local;
        when c_L0_counter       => data_out <= X"0000" & L0_Counter;
        when c_L1_counter       => data_out <= X"0000" & L1_Counter;
        when c_L1_msg_counter   => data_out <= X"0000" & L1_msg_Counter;
        when c_L2a_counter      => data_out <= X"0000" & L2a_Counter;
        when c_L2r_counter      => data_out <= X"0000" & L2r_Counter;
        when c_hammingerrorcnt  => data_out <= hammingErrorCnt;
        when c_errorcnt         => data_out <= seqErrorCnt & msgErrorCnt;
        when c_buffered_events  => data_out <= X"0000000" & buffered_events;
        when c_DAQ_header01     => data_out <= DAQ_header(0);
        when c_DAQ_header02     => data_out <= DAQ_header(1);
        when c_DAQ_header03     => data_out <= DAQ_header(2);
        when c_DAQ_header04     => data_out <= DAQ_header(3);
        when c_DAQ_header05     => data_out <= DAQ_header(4);
        when c_DAQ_header06     => data_out <= DAQ_header(5);
        when c_DAQ_header07     => data_out <= DAQ_header(6);
        when c_event_info       => data_out <= X"000" & "00" & event_info_reg;  
        when c_event_error      => data_out <= X"0" & "000" & event_error_reg;    
        when others             => data_out <= (others => '-');
      end case;
    elsif (we='1' and rcu_addr(15 downto 12) = c_Module_address) then --writing
      case rcu_addr(11 downto 0) is
        when c_control =>
          input_enable_i       <= data_in(0);
          disable_error_mask_i <= data_in(1);
          L0_support_i         <= data_in(3);
          L2a_mask_i           <= data_in(8);
          L2r_mask_i           <= data_in(9);
          L2_timeout_mask_i    <= data_in(10);
          L1a_msg_mask_i       <= data_in(11);
          trigger_input_mask_enable_i    <= data_in(12);
          
        when c_module_reset =>
          module_reset         <= '1';
        when c_Reset_Counters =>
          counter_reset        <= '1';
        when c_L1_Latency =>
          L1_boundary_reg      <= data_in(15 downto 12);
          L1_Latency_reg       <= data_in(11 downto 0); 
        when c_L2_Latency =>
          L2_Latency_reg       <= data_in;
        when c_L2_ExtendedLatency =>
          L2_ExtendedLatency_reg <= data_in;
        when c_L1_msg_Latency =>
          L1_msg_Latency_reg   <= data_in;
        when others =>
      end case;
    end if;
  end if;
end process p_rcu_communication;


end arc;
