-------------------------------------------------------------------------------
-- Title      : Sequence Validator
-- Project    : RCU Trigger Receiver
-------------------------------------------------------------------------------
-- File             : $RCSfile: sequence_validator.vhd,v $
-- Last edited by   : $Author: alme $
-- Last update      : $Date: 2010-03-24 07:30:25 $
-- Current Revision : $Revision: 1.16 $
-------------------------------------------------------------------------------
-- Description: Validates the correct order and timing of the trigger 
-- and messages
-------------------------------------------------------------------------------
-- Revision History : 
-- http://www.ift.uib.no/kjekscgi-bin/viewcvs.cgi/vhdlcvs/trigger_receiver/
-------------------------------------------------------------------------------
--
-- This file is property of and copyright by the Instrumentation and 
-- Electronics Section, Dep. of Physics and Technology
-- University of Bergen, Norway, 2005
-- This file has been written by Johan Alme
-- Johan.Alme@ift.uib.no
--
-- Permission to use, copy, modify and distribute this firmware and its  
-- documentation strictly for non-commercial purposes is hereby granted  
-- without fee, provided that the above copyright notice appears in all  
-- copies and that both the copyright notice and this permission notice  
-- appear in the supporting documentation. The authors make no claims    
-- about the suitability of this firmware for any purpose. It is         
-- provided "as is" without express or implied warranty.                 
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.trigger_receiver_pkg.all;

entity sequence_validator is
  port(
    clk                 : in  std_logic;
    reset_n             : in  std_logic;
    module_reset        : in  std_logic;
    -- triggers in
    pre_pulse           : in  std_logic;
    L0                  : in  std_logic;
    L1                  : in  std_logic;
    -- receiving messages
    receiving_L1        : in  std_logic;
    receiving_L2a       : in  std_logic;
    receiving_L2r       : in  std_logic;
    L1a_msg_mask        : in  std_logic;
    -- data used for error checking and status bits
    BCID_L2a            : in  std_logic_vector(11 downto 0);
    BCID_L2r            : in  std_logic_vector(11 downto 0);
    ESR_L1              : in  std_logic;
    ESR_L2a             : in  std_logic;
    L1_SwC              : in  std_logic;
    L2_SwC              : in  std_logic;
    CIT_L1              : in  std_logic;
    CIT_L2a             : in  std_logic;
    RoC                 : in  std_logic_vector(3 downto 0);
    -- Legal time windows.
    L0_tw               : in  std_logic; -- only applicable when pre-pulse
    L1_tw               : in  std_logic;
    L1_exact_tw         : in  std_logic;
    L1_msg_tw           : in  std_logic;
    L2_tw               : in  std_logic;
    L1_tw_passed        : in  std_logic;
    L1_msg_tw_passed    : in  std_logic;
    L2_tw_passed        : in  std_logic;
    -- outputs
    info_reg            : out std_logic_vector(12 downto 0);
    error_reg           : out std_logic_vector(14 downto 0);
    init_sequence       : out std_logic; -- used to reset all status/error bits
    busy                : out std_logic;
    sequence_error      : out std_logic; -- pulse to counters when error.
    start_of_run        : out std_logic;
    end_of_run          : out std_logic;
    sync                : out std_logic;
    run_active          : out std_logic;
    L1r_out             : out std_logic;
    L1_msg_out          : out std_logic;
    L2a_out             : out std_logic;
    L2r_out             : out std_logic;
    L2_timeout          : out std_logic    
  );
end sequence_validator;

architecture behave of sequence_validator is

  type state is (s_idle,              -- default state
                 s_wait,
                 s_wait_for_L1_tw,    -- legal time window for L1
                 s_wait_for_L1,       -- waiting for L1
                 s_issue_L1r,
                 s_receive_messages,
                 s_L2_timeout,
                 s_L2a,
                 s_L2r);
  signal current_state, next_state : state;
  
  signal receiving_L1R       : std_logic;
  signal receiving_L2aR      : std_logic;
  signal receiving_L2rR      : std_logic;
  signal L2_twR              : std_logic;
  signal receiving_messages  : std_logic;
  
  signal L1_msg_end          : std_logic;
  signal L2a_end             : std_logic;
  signal L2r_end             : std_logic;
  signal L2_tw_end           : std_logic;    
  
  signal RoI_announced    : std_logic;                    -- info_reg(1) = ESR_L1
  signal sw_trigger_type  : std_logic_vector(3 downto 0); -- info_reg(7 downto 4) = ROC
  signal cal_trigger      : std_logic;                    -- info_reg(8)  = pre-pulse
  signal sw_trigger       : std_logic;                    -- info_reg(9)  = L1_SwC
  signal L2a_received     : std_logic;                    -- info_reg(10)
  signal L2r_received     : std_logic;                    -- info_reg(11)
  signal payload_on       : std_logic;                    -- info_reg(12)
  
  signal pre_pulseR       : std_logic;
  signal timeout_counter  : std_logic_vector(3 downto 0):=X"0";
  
  signal spurious_L0         : std_logic; -- error_reg(0)
  signal missing_L0          : std_logic; -- error_reg(1)
  signal spurious_L1         : std_logic; -- error_reg(2)
  signal boundary_L1         : std_logic; -- error_reg(3)
  signal missing_L1          : std_logic; -- error_reg(4)
  signal spurious_L1_msg     : std_logic; -- error_reg(5) timing violation
  signal missing_L1_msg      : std_logic; -- error_reg(6)
  signal spurious_L2_msg     : std_logic; -- error_reg(7) timing violation
  signal missing_L2_msg      : std_logic; -- error_reg(8)
  signal prepulse_error      : std_logic; -- error_reg(12) added in case of future implementation
  signal L1_msg_content_err  : std_logic; -- error_reg(13)
  signal L2_msg_content_err  : std_logic; -- error_reg(14)
  
begin
  
  -- assigning outputs
  info_reg     <= payload_on & L2a_received & L2r_received & cal_trigger & sw_trigger & sw_trigger_type & "00" & RoI_announced & '0';
  --masking L1 message errors if L1 message not chosen to be retrived.
  error_reg    <= '0' & L2_msg_content_err & (L1_msg_content_err and L1a_msg_mask) & prepulse_error &
                  '0' & '0' & missing_L2_msg & spurious_L2_msg & (missing_L1_msg and L1a_msg_mask) & spurious_L1_msg &
                  missing_L1 & boundary_L1 & spurious_L1 & missing_L0 & spurious_L0;

  payload_on <= (L2a_received and not missing_L1 and not missing_L0);

  --purpose: extra register to check whether 
  p_edge : process(clk)
  begin
    if rising_edge(clk) then
      receiving_L1R  <= receiving_L1;
      receiving_L2aR <= receiving_L2a;
      receiving_L2rR <= receiving_L2r;
      L2_twR         <= L2_tw;
    end if;
  end process p_edge;
  
  receiving_messages <= receiving_L1 or receiving_L2a or receiving_L2r;

  -- detect end of messages transmissions
  L1_msg_end <= '1' when (receiving_L1R = '1'  and receiving_L1 = '0')  else '0';
  L2r_end    <= '1' when (receiving_L2rR = '1' and receiving_L2r = '0') else '0';
  L2a_end    <= '1' when (receiving_L2aR = '1' and receiving_L2a = '0') else '0';
  L2_tw_end  <= '1' when (L2_twR = '1'         and L2_tw = '0')         else '0';
  
  -- Purpose: Sets current state.
  p_sm_driver : process(clk, reset_n)
  begin
    if (reset_n = '0') then
      current_state <= s_idle;
    elsif rising_edge(clk) then
      if (module_reset = '1') then
        current_state <= s_idle;
      else
        current_state <= next_state;
      end if;
    end if;    
  end process p_sm_driver;

  

  -- Purpose: sets next state of state machine. In short a valid sequence is expected to be
  -- L0-L1-L1msg-(RoI)-L2msg (l2a/L2r) for a normal data event or a software trigger.
  -- A calibration sequence is expected to start with a pre-pulse.
  p_next_state : process(current_state, L0, L1, L1_tw, L1_tw_passed, L2_tw, L2_tw_passed, 
                    receiving_messages, L2a_end, L2r_end, L2_tw_end, timeout_counter)
  begin
    case current_state is
      when s_idle =>
        if (L0 = '1') then -- normal trigger/software trigger
          next_state <= s_wait;
        elsif (L1 = '1') then -- missing L0
          next_state <= s_receive_messages;
        elsif (receiving_messages = '1') then
          next_state <= s_receive_messages; -- missing all triggers
        else 
          next_state <= s_idle;
        end if;
      
      when s_wait =>
        if L1_tw = '1' then -- not reset from default high
          next_state <= s_wait;
        else
          next_state <= s_wait_for_L1_tw;
        end if;
      -- reception of Level 1 trigger
      when s_wait_for_L1_tw =>
        if (L1_tw_passed = '1') then
          next_state <= s_wait_for_L1;
        elsif (L1_tw = '1') then
          if (L1 = '1') then -- in case L1 arrives when tw goes high. 
            next_state <= s_receive_messages;
          else
            next_state <= s_wait_for_L1;
          end if;
        else
          next_state <= s_wait_for_L1_tw;
        end if;
      when s_wait_for_L1 =>
        if (L1_tw = '1' and L1 = '1') then -- valid L1
          next_state <= s_receive_messages;
        elsif (L1_tw = '0') then -- timeout / L1 missing
          next_state <= s_issue_L1r; -- valid L0 sequence
        else 
          next_state <= s_wait_for_L1;
        end if;
      
      when s_issue_L1r =>
        next_state <= s_idle;
      
      -- message reception
      when s_receive_messages =>
        if (L2_tw_end = '1' or L2_tw_passed = '1') then -- end of time window - end of sequence - timeout
          next_state <= s_L2_timeout;
        elsif (L2_tw = '1') then
          if (L2a_end = '1') then --finished receiving L2a message
            next_state <= s_L2a;
          elsif (L2r_end = '1') then -- finished receiving L2r message
            next_state <= s_L2r;
          else
            next_state <= s_receive_messages;
          end if;
        else 
          next_state <= s_receive_messages;
        end if;
      
      when s_L2_timeout | s_L2a | s_L2r =>
        next_state <= s_idle;
        
      when others =>
        next_state <= s_idle;
    end case;
  end process p_next_state;           

  -- Purpose: Sets all outputs of the state machine. This be triggers, errors and special messages decing (etc. run active)
  p_set_output : process(clk, reset_n)
  begin
    if (reset_n = '0') then
      init_sequence       <= '0';
      busy                <= '0';
      pre_pulseR          <= '0';
      run_active          <= '0';
      sync                <= '0'; 
      start_of_run        <= '0';
      end_of_run          <= '0';
      L1r_out             <= '0';
      L1_msg_out          <= '0';
      L2a_out             <= '0';
      L2r_out             <= '0';
      L2_timeout          <= '0';
      L2a_received        <= '0';
      L2r_received        <= '0';
      --info  
      RoI_announced       <= '0';
      sw_trigger_type     <= (others => '0');
      cal_trigger         <= '0';
      sw_trigger          <= '0';
      -- error
      sequence_error      <= '0';   --pulse for all errors
      spurious_L0         <= '0';
      missing_L0          <= '0';
      spurious_L1         <= '0';
      missing_L1          <= '0';
      boundary_L1         <= '0';
      spurious_L1_msg     <= '0';
      missing_L1_msg      <= '0';
      spurious_L2_msg     <= '0';
      missing_L2_msg      <= '0';
      prepulse_error      <= '0';
      L1_msg_content_err  <= '0';
      L2_msg_content_err  <= '0';
      timeout_counter     <= X"0";
      
    elsif rising_edge(clk) then
      --defaults
      sequence_error      <= '0';   --pulse for all errors
      init_sequence       <= '0';
      timeout_counter     <= X"0";
      L1r_out             <= '0';
      L1_msg_out          <= '0';
      L2a_out             <= '0';
      L2r_out             <= '0';
      L2_timeout          <= '0';
      sync                <= '0'; 
      start_of_run        <= '0';
      end_of_run          <= '0';
      
      if (L0 = '1' and L0_tw = '0') then
        spurious_L0         <= '1';
        sequence_error      <= '1';
      end if;
      if (L1 = '1' and L1_tw = '1' and L1_exact_tw = '0') then
        boundary_L1         <= '1';
        sequence_error      <= '1';
      end if;
      if (L1 = '1' and L1_tw = '0') then
        spurious_L1         <= '1';
        sequence_error      <= '1';
      end if;
      if (receiving_L1 = '1' and L1_msg_tw = '0' and spurious_L1_msg = '0') then--only count spurious message when it occurs
        spurious_L1_msg     <= '1';
        sequence_error      <= '1';
      end if;
      if ((receiving_L2a = '1' or receiving_L2r = '1') and L2_tw = '0' and spurious_L2_msg = '0') then
        spurious_L2_msg     <= '1';
        sequence_error      <= '1';
      end if;

      case current_state is 
        when s_idle =>
          busy <= '0';
          if (pre_pulse = '1') then
            pre_pulseR          <= '1';
          end if;      
          if (L0 = '1' or L1 = '1' or module_reset = '1' or receiving_L1 = '1' or receiving_L2a = '1' or receiving_L2r = '1') then
            init_sequence       <= L0 or L1;  -- reset other error/info regs only if sequence starts with trigger.
			
            busy                <= '1'; -- set busy flag until L2 or L2 timeout.
            -- reseting error/info from previous trigger
            L2a_received        <= '0';
            L2r_received        <= '0';
            RoI_announced       <= '0';
            sw_trigger          <= '0';
            sw_trigger_type     <= (others => '0');
            spurious_L0         <= '0';
            spurious_L1         <= '0';
            missing_L0          <= '0';
            missing_L1          <= '0';
            boundary_L1         <= '0';
            spurious_L1_msg     <= '0';
            missing_L1_msg      <= not module_reset;--making sure that these are zero when doing module reset
            spurious_L2_msg     <= '0';
            missing_L2_msg      <= not module_reset;
            prepulse_error      <= '0';
            L1_msg_content_err  <= '0';
            L2_msg_content_err  <= '0';
            sequence_error      <= '0';
            pre_pulseR          <= '0' ; --reset pre pulse flag...
            
            -- detect sequence errors
            if (L1 = '1') then
              missing_L0          <= '1';
              sequence_error      <= '1';
            end if;
            
            if (receiving_messages = '1') then
              missing_L0          <= '1';
              missing_L1          <= '1';
              sequence_error      <= '1';
            end if;
         end if;
        
        when s_wait =>
          -- do nothing
          
        -- reception of Level 1 trigger
        when s_wait_for_L1_tw =>
          
        when s_wait_for_L1 =>
          if (L1_tw = '0') then
            missing_L1          <= '1';
          end if;
          
        when s_issue_L1r =>
          L1r_out <= '1';
          
        when s_receive_messages =>
          --verifiying L1 message
          if (L1_msg_end = '1' and L1_msg_tw = '1') then
            
            -- ERROR CHECKING OF L1 MESSAGE
            if (L1_SwC = '0' and CIT_L1 = '1') then        -- a calibration trigger is also a software trigger
              L1_msg_content_err   <= '1';
              sequence_error       <= '1';
            end if;
           
            --- DONE ERROR CHECKING

            -----------------------------------------------------------------------------------------
            RoI_announced        <= ESR_L1;
            sw_trigger           <= L1_SwC;
            sw_trigger_type      <= RoC;
            missing_L1_msg       <= '0';
            L1_msg_out           <= '1';
            
          end if;
          
        when s_L2a =>
          if (missing_L1_msg = '1') then
            sequence_error       <= '1';
          end if;
          
          -- ERROR CHECKING OF L2A MESSAGE
            
          if (L2_SwC = '0' and CIT_L2a = '1') then        -- a calibration trigger is also a software trigger
            L2_msg_content_err   <= '1';
            sequence_error       <= '1';
          end if; 
             
          -- BCID should not be larger than 3563 (out of range) (ref. fe_error.doc 19.10.2006)
          if (BCID_L2a > X"DEB") then
            L2_msg_content_err    <= '1';
            sequence_error        <= '1';
          end if;    

          if (L1_SwC /= L2_SwC or ESR_L1 /= ESR_L2a or CIT_L1 /= CIT_L2a) then -- or cal_trigger /= CIT_L2a) then
            sequence_error     <= '1';
            L2_msg_content_err <= '1';
          end if;
          
          -- DONE ERROR CHECKING L2a message
          
          -----------------------------------------------------------------------------------------
          -- Here's the place to decode the messages, for example different SW triggers etc.
          -----------------------------------------------------------------------------------------
          -- setting cal trigger
          cal_trigger <= CIT_L1 or CIT_L2a; 

          -- setting active run based on Start of Run trigger and End of Run trigger
          if (L1_SwC = '1' and L2_SwC = '1') then -- sw trigger arrived
            if (RoC = X"D") then -- start of Run
              sync         <= '1'; 
            elsif (RoC = X"E") then -- start of Run
              run_active   <= '1';
              start_of_run <= '1'; 
            elsif (RoC = X"F") then --end of Run
              run_active <= '0';
              end_of_run <= '1'; 
            end if; 
          end if;
             
          -----------------------------------------------------------------------------------------
          busy                <= '0'; -- sequence done
          missing_L2_msg      <= '0'; --timeout
          L2a_received        <= '1';
          L2a_out             <= '1';
          
        when s_L2r =>
          if (missing_L1_msg = '1') then
            sequence_error       <= '1';
          end if;
          -- ERROR CHECKING OF L2R MESSAGE
          -- BCID should not be larger than 3563 (out of range) (ref. fe_error.doc 19.10.2006)
          if (BCID_L2r > X"DEB") then
            L2_msg_content_err    <= '1';
          end if; 
            
          -----------------------------------------------------------------------------------------
          busy                <= '0'; -- sequence done
          missing_L2_msg      <= '0'; -- not timeout
          L2r_received        <= '1';
          L2r_out             <= '1';
        
        when s_L2_timeout =>
          if (missing_L1_msg = '1') then
            sequence_error       <= '1';
          end if;
          
          busy                <= '0';     -- sequence timedout.
          L2_timeout          <= '1';
          sequence_error      <= '1';        
        
        when others =>
          -- do nothing;
      end case;
    end if;    
  end process p_set_output;

end architecture behave;  
