-------------------------------------------------------------------------------
-- Title      : Serial B Communication Module
-- Project    : RCU Trigger Receiver
-------------------------------------------------------------------------------
-- File             : $RCSfile: serialb_com.vhd,v $
-- Last edited by   : $Author: alme $
-- Last update      : $Date: 2007/12/10 15:26:24 $
-- Current Revision : $Revision: 1.9 $
-------------------------------------------------------------------------------
-- Description: Deserializer for the Serial B line.
-------------------------------------------------------------------------------
-- Revision History : 
-- http://www.ift.uib.no/kjekscgi-bin/viewcvs.cgi/vhdlcvs/trigger_receiver/
-------------------------------------------------------------------------------
--
-- This file is property of and copyright by the Instrumentation and 
-- Electronics Section, Dep. of Physics and Technology
-- University of Bergen, Norway, 2005
-- This file has been written by Johan Alme
-- Johan.Alme@ift.uib.no
--
-- Permission to use, copy, modify and distribute this firmware and its  
-- documentation strictly for non-commercial purposes is hereby granted  
-- without fee, provided that the above copyright notice appears in all  
-- copies and that both the copyright notice and this permission notice
-- appear in the supporting documentation. The authors make no claims
-- about the suitability of this firmware for any purpose. It is
-- provided "as is" without express or implied warranty.                 
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.trigger_receiver_pkg.all;

entity serialb_com is
  port(
    clk                 : in  std_logic;
    reset_n             : in  std_logic;
    module_reset        : in  std_logic;
    serialBchannel      : in  std_logic;
    init_sequence       : in  std_logic;
    error_reg           : out std_logic;
    brcdata_ready       : out std_logic;
    adrdata_ready       : out std_logic;
    received_data       : out std_logic_vector(38 downto 0)
  );
end serialb_com;

architecture behave of serialb_com is

  type state is (s_idle, s_fmt, s_get_addressed_data, s_end_addressed_data, 
                 s_get_broadcast_data, s_end_broadcast_data, s_wait);

  signal received_word        : std_logic_vector(38 downto 0);
  signal stop_bit_error_i     : std_logic;
  signal serialBchannelR      : std_logic;
  signal serialBchannelRR     : std_logic;

  signal current_state        : state;    
begin

  received_data <= received_word;

-- purpose: receiving serial B messages by shifting them into a shift register.
--          checking the hamming bits.
p_driver: process(clk, reset_n)
begin
  if (reset_n = '0') then
    current_state              <= s_idle;
    received_word(0)           <= '1';
    received_word(38 downto 1) <= (others => '0');
    stop_bit_error_i           <= '0';
    brcdata_ready              <= '0';
    adrdata_ready              <= '0';
    
  elsif rising_edge(clk) then
    -- assign default values
    current_state              <= s_idle;
    --shift reg.
    stop_bit_error_i           <= '0';
    brcdata_ready              <= '0';
    adrdata_ready              <= '0';
    
    -- always return to idle when doing module reset.
    if (module_reset = '1') then
      current_state <= s_idle;
    end if;
    
    
    case current_state is
      when s_idle =>
       	if (serialBchannelRR = '0') then --start bit
       	  -- init shift reg
       	  received_word(0)           <= '1';
          received_word(38 downto 1) <= (others => '0');
          --set next state.
          current_state <= s_fmt;
        else
          current_state <= s_idle;
        end if;
      
      when s_fmt =>
        if (serialBchannelRR = '1') then  -- fmt bit is 1 for ttcrx addr receive.
          current_state <= s_get_addressed_data;
        else                        -- fmt bit is 0 for ttcrx broadcast
          current_state <= s_get_broadcast_data;
        end if;
      
      when s_get_addressed_data => -- shiftregister for serialB input
        received_word(38 downto 1) <= received_word(37 downto 0);
        received_word(0)           <= serialBchannelRR;
      
        if (received_word(38) = '1') then
          current_state       <= s_end_addressed_data;
        else
          current_state       <= s_get_addressed_data;
        end if;
      
      when s_get_broadcast_data => -- shiftregister for serialB input
        received_word(38 downto 1) <= received_word(37 downto 0);
        received_word(0)           <= serialBchannelRR;
  
        if (received_word(12) = '1') then
          current_state       <= s_end_broadcast_data;
        else
          current_state       <= s_get_broadcast_data;
        end if;
      
      when s_end_addressed_data => 
        stop_bit_error_i <= not serialBChannelRR; -- stop bit should be 1
        
        current_state <= s_wait;
        adrdata_ready <= '1';
    
      when s_end_broadcast_data => 
        stop_bit_error_i <= not serialBChannelRR; -- stop bit should be 1
        
        current_state <= s_wait;
        brcdata_ready <= '1';
    
      when s_wait =>
        current_state <= s_idle;
      
      when others =>
    	   current_state <= s_idle;
    end case;
  end if;
end process p_driver;

-- purpose: synchronizing the serial B data line, and turning off the serial B data line.
p_synch1: process(clk)
begin
  if (falling_edge(clk)) then
    serialBchannelR     <= serialBchannel;
  end if;  
end process p_synch1;

-- purpose: synchronizing the serial B data line, and turning off the serial B data line.
p_synch2: process(clk)
begin
  if (rising_edge(clk)) then
    serialBchannelRR	  <= serialBchannelR;
  end if;  
end process p_synch2;


-- purpose: storing the error.
p_error_reg: process(clk, reset_n)
begin
  if (reset_n='0') then
    error_reg   <= '0';
  elsif (rising_edge(clk)) then
    if (init_sequence = '1' or module_reset = '1') then
      error_reg <= '0';
    elsif (stop_bit_error_i = '1') then
      error_reg <= '1';
    end if;
  end if;  
end process p_error_Reg;


end architecture behave;

