-------------------------------------------------------------------------------
-- Title      : Trigger Receiver Module
-- Project    : RCU Trigger Receiver
-------------------------------------------------------------------------------
-- File             : $RCSfile: trigger_receiver.vhd,v $
-- Last edited by   : $Author: alme $
-- Last update      : $Date: 2009-10-19 05:13:57 $
-- Current Revision : $Revision: 1.18 $
-------------------------------------------------------------------------------
-- Description      : Receives and decodes data from the Trigger system using
--                    the serial B channel and channel A that has L0 and L1a
--                    triggers. Checks the validity of the sequence, and store
--                    the data as DAQ Header words (including info/error)
--                    in a FIFO. 
-------------------------------------------------------------------------------
-- Revision History : 
-- http://www.ift.uib.no/kjekscgi-bin/viewcvs.cgi/vhdlcvs/trigger_receiver/
-------------------------------------------------------------------------------
--
-- This file is property of and copyright by the Instrumentation and 
-- Electronics Section, Dep. of Physics and Technology
-- University of Bergen, Norway, 2005
-- This file has been written by Johan Alme
-- Johan.Alme@ift.uib.no
--
-- Permission to use, copy, modify and distribute this firmware and its  
-- documentation strictly for non-commercial purposes is hereby granted  
-- without fee, provided that the above copyright notice appears in all  
-- copies and that both the copyright notice and this permission notice  
-- appear in the supporting documentation. The authors make no claims    
-- about the suitability of this firmware for any purpose. It is         
-- provided "as is" without express or implied warranty.                 
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.trigger_receiver_pkg.all;

entity trigger_receiver is
  generic (
    include_fifo_parity     : integer := 1;    -- generates parity bit for data written to the FIFO
	platform				: integer := 1;
    include_phase_check     : integer := 1);
  port(
    -- general signals
    clk               : in    std_logic;                    -- global clock 40 MHz
    sclk              : in    std_logic;
    reset_n           : in    std_logic;                    -- global asynch reset from DCS board
    soft_reset        : in    std_logic;                    -- synch soft reset to reset status
    -- ttc communication
    L1Accept          : in    std_logic;                    -- input from TTCrx with L0 and L1 trigger
    serialBchannel    : in    std_logic;                    -- input from TTCrx with message transmissions
    -- outputs
    FEE_reset         : out   std_logic;
    busy              : out   std_logic;                    -- high during reception of sequence.
    cal_pre_pulse     : out   std_logic;                    -- decoded pre-pulse broadcast message
    start_of_run      : out   std_logic;                    -- pulse when start of run event is decoded.
    end_of_run        : out   std_logic;                    -- pulse when end of run event is decoded.
    sync              : out   std_logic;                    -- pulse when receiving a sync software trigger
    event_reset       : out   std_logic;                    -- decoded event reset broadcast message
    bunch_reset       : out   std_logic;                    -- decoded bunch reset broadcast message
    bunchcnt_overflow : out   std_logic;                    -- bunchcounter overflow flag.
    L0_trigger        : out   std_logic;                    -- decoded L0 trigger from L1Accept line
    L1a_trigger       : out   std_logic;                    -- decoded L1a trigger from L1Accept line
    L1r_trigger       : out   std_logic;                    -- L1 reject trigger - to be sent when no L1 arrives in correct time.
    L2a_trigger       : out   std_logic;                    -- decoded L2a trigger from serialB message reception
    L2r_trigger       : out   std_logic;                    -- decoded L2r trigger from serialB message reception
    L2_timeout        : out   std_logic;                    -- Pulse if sequence is ended with missing L2 message
    DAQ_header_out    : out   std_logic_vector(32 downto 0);-- Output from FIFO with event info and DAQ headers
    read_counter_out  : out   std_logic_vector(3 downto 0);
    buffered_events   : out   std_logic_vector(3 downto 0); -- Number of events stored in FIFO at any given time
    -- inputs from internal logic
    detector          : in    std_logic;                    -- 0 = TPC, 1 = PHOS
    read_enable       : in    std_logic;                    -- Input to read FIFO.
    meb_full          : in    std_logic;
    ddl_rdy_rx        : in    std_logic;
    -- rcu bus interface
    data_in           : in    std_logic_vector(31 downto 0);-- RCU bus data in
    data_out          : out   std_logic_vector(31 downto 0);-- RCU bus data out
    RCU_addr          : in    std_logic_vector(15 downto 0);-- Address from DCS board
    we                : in    std_logic                     -- Write enable, postive polarity
  );
end trigger_receiver;

-- Description: The top level architecturs contains mostly only sub modules.
-- Some logic is added to generate DAQ headers from the ifnormation of the sub modules etc
architecture arc of trigger_receiver is

  signal serialBchannel_i    : std_logic; -- top level --> serialb_com
  signal L1Accept_i          : std_logic; -- top level --> serialb_com
  signal input_enable        : std_logic; -- rcu_com --> top level
  signal serialB_error_reg   : std_logic; -- serialb_com --> rcu com / Fifo
  signal brcdata_ready       : std_logic; -- serialb_com --> hamming_decoder
  signal adrdata_ready       : std_logic; -- serialb_com --> hamming_decoder
  signal received_data       : std_logic_vector(38 downto 0); -- serialb_com --> hamming decoder
  signal single_bit_error    : std_logic; -- hamming_decoder --> counters / error reg
  signal double_bit_error    : std_logic; -- hamming_decoder --> counters / error reg
  signal hmg_error_reg       : std_logic_vector(3 downto 0); -- hamming_decoder --> rcu com / Fifo
  signal hmg_brcdata_ready   : std_logic; -- hamming_decoder --> broadcast_msg_decoder
  signal hmg_adrdata_ready   : std_logic; -- hamming_decoder --> addressed_msg_decoder
  signal brcdata             : std_logic_vector(7  downto 0); -- hamming_decoder --> broadcast_msg_decoder
  signal adrdata             : std_logic_vector(31 downto 0); -- hamming_decoder --> addressed_msg_decoder
  signal bunchCnt_reset      : std_logic; -- broadcast_msg_decoder --> Output / counters
  signal eventCnt_reset      : std_logic; -- broadcast_msg_decoder --> Output
  signal pre_pulse           : std_logic; -- broadcast_msg_decoder --> Output / sequence validator
  signal init_sequence       : std_logic; -- sequence_validator --> addressed_message_decoder / hamming decoder
  signal msg_error_reg       : std_logic_vector(4 downto 0); -- addressed_message_decoder --> FIFO /RCU com
  signal msg_error           : std_logic; -- addressed_message_decoder --> counters
  signal receiving_L1        : std_logic; -- addressed_message_decoder --> sequence_validator
  signal receiving_L2a       : std_logic; -- addressed_message_decoder --> sequence_validator
  signal receiving_L2r       : std_logic; -- addressed_message_decoder --> sequence_validator
  signal receiving_L2        : std_logic; 
  signal L1_message          : message_array(0 to 0); -- addressed_message_decoder --> RCU_com / FIFO
  signal L2a_message         : message_array(0 to 7); -- addressed_message_decoder --> RCU_com / FIFO
  signal L2r_message         : message_array(0 to 0); -- addressed_message_decoder --> RCU_com / FIFO
  signal L0                  : std_logic; -- L1 Line Decoder --> Sequence Validator / Output
  signal L1a                 : std_logic; -- L1 Line Decoder --> Sequence Validator / Output
  signal BCID_local          : std_logic_vector(11 downto 0); -- counters --> sequence validator / rcu com
  signal L2a_mask            : std_logic; -- rcu com --> top level
  signal L2r_mask            : std_logic; -- rcu com --> top level
  signal L2_timeout_mask     : std_logic; -- rcu com --> top level

  signal L0_tw               : std_logic; -- counters --> sequence validator
  signal L1_tw               : std_logic; -- counters --> sequence validator
  signal L1_exact_tw         : std_logic;
  signal L1_msg_tw           : std_logic; -- counters --> sequence validator
  signal L2_tw               : std_logic; -- counters --> sequence validator
  signal L0_masked_tr        : std_logic; -- Masked triggers to remove spurious on output
  signal L1a_masked_tr       : std_logic; -- Masked triggers to remove spurious on output
  signal L2a_masked_tr       : std_logic; -- Masked triggers to remove spurious on output
  signal L2r_masked_tr       : std_logic; -- Masked triggers to remove spurious on output
  signal L1_tw_passed        : std_logic; -- counters --> sequence validator
  signal L1_msg_tw_passed    : std_logic; -- counters --> sequence validator
  signal L2_tw_passed        : std_logic; -- counters --> sequence validator
  signal seq_info_reg        : std_logic_vector(12 downto 0);  -- sequence validator --> rcu com / FIFO
  signal seq_error_reg       : std_logic_vector(14 downto 0); -- sequence validator --> rcu com / FIFO
  signal sequence_error      : std_logic; -- sequence validator --> counters
  signal run_active_i        : std_logic; -- sequence validator --> output
  signal L1r_tr              : std_logic;
  signal L1_msg_tr           : std_logic; -- sequence validator --> counters
  signal L2a_tr              : std_logic; -- sequence validator --> counters / output
  signal L2r_tr              : std_logic; -- sequence validator --> counters / output
  signal L2_timeout_i        : std_logic; -- sequence validator --> output
  signal counter_reset       : std_logic; -- rcu com --> counters
  signal L1_Latency          : std_logic_vector(11 downto 0); -- rcu com --> counters
  signal L1_boundary         : std_logic_vector( 3 downto 0);
  signal L2_Latency          : std_logic_vector(31 downto 0); -- rcu com --> counters
  signal L1_msg_Latency      : std_logic_vector(31 downto 0); -- rcu com --> counters
  signal Bunch_Counter       : std_logic_vector(11 downto 0); -- counters --> rcu com
  signal BunchCnt_overflow_i : std_logic; -- counters --> rcu com
  signal pp_counter          : std_logic_vector(15 downto 0); -- counters --> rcu com
  signal L0_Counter          : std_logic_vector(15 downto 0); -- counters --> rcu com
  signal L1_Counter          : std_logic_vector(15 downto 0); -- counters --> rcu com
  signal L1_msg_Counter      : std_logic_vector(15 downto 0); -- counters --> rcu com
  signal L2a_Counter         : std_logic_vector(15 downto 0); -- counters --> rcu com
  signal L2r_Counter         : std_logic_vector(15 downto 0); -- counters --> rcu com
  signal hammingError_Counter: std_logic_vector(31 downto 0); -- counters --> rcu com
  signal msgError_Counter    : std_logic_vector(15 downto 0); -- counters --> rcu com
  signal seqError_Counter    : std_logic_vector(15 downto 0); -- counters --> rcu com
  signal module_reset_i      : std_logic; -- rcu com --> all submodules
  signal module_reset        : std_logic; -- rcu com --> all submodules
  signal header_version      : std_logic_vector(3 downto 0); -- rcu com --> FIFO
  signal buffered_events_i   : std_logic_vector(3 downto 0); -- FIFO --> rcu com/output
  signal event_info_reg      : std_logic_vector(31 downto 0); -- top level --> rcu_com/FIFO
  signal event_error_reg     : std_logic_vector(31 downto 0); -- top level --> rcu_com/FIFO
  signal DAQ_header          : DAQ_header_array(0 to 6); -- top level --> rcu_com/FIFO
  signal L2_trigger          : std_logic; -- L2r/L2a/L2_timeout --> addressed messagedecoder/FIFO
  signal read_enable_i       : std_logic; -- input --> FIFO
  signal DAQ_headerFIFO      : std_logic_vector(32 downto 0); --FIFO --> output
  signal busy_i              : std_logic; -- sequence_validator --> rcu com, counters, output
  alias BCID_L2a             : std_logic_vector(11 downto 0) is L2a_message(0);
  alias BCID_L2r             : std_logic_vector(11 downto 0) is L2r_message(0);
  alias ESR_L1               : std_logic is L1_message(0)(5);
  alias ESR_L2a              : std_logic is L2a_message(3)(10);
  alias L1_SwC               : std_logic is L1_message(0)(4);
  alias L2_SwC               : std_logic is L2a_message(3)(8);
  alias CIT_L1               : std_logic is L1_message(0)(10);
  alias CIT_L2a              : std_logic is L2a_message(3)(9);
  alias RoC                  : std_logic_vector(3 downto 0) is L1_message(0)(9 downto 6);
  signal OrbitID             : std_logic_vector(23 downto 0);
  signal L2Class             : std_logic_vector(49 downto 0);
  signal DAQ_errorStatus     : std_logic_vector(15 downto 0);
  signal BCID                : std_logic_vector(11 downto 0);
  signal trigger             : std_logic;
  signal phase               : std_logic_vector(4 downto 0);
  signal L0_support          : std_logic;
  signal L1a_msg_mask        : std_logic;
  signal disable_error_mask  : std_logic; 
  signal trigger_input_mask  : std_logic;
  signal trigger_input_mask_enable : std_logic;  
  signal soft_reset_i        : std_logic;

begin

  -- assigning outputs
  event_reset       <= eventCnt_reset;
  bunch_reset       <= bunchCnt_reset;
  buffered_events   <= buffered_events_i;
  cal_pre_pulse     <= pre_pulse;
  busy              <= busy_i;
  bunchcnt_overflow <= BunchCnt_overflow_i;
  L0_masked_tr      <= L0 and L0_tw      when trigger_input_mask = '0' else '0'; --removing spurious triggers and masking inputs if no more triggers can be accepted
  L1a_masked_tr     <= L1a and L1_tw     when trigger_input_mask = '0' else '0';
  L2a_masked_tr     <= L2a_tr and L2_tw  when trigger_input_mask = '0' else '0';
  L2r_masked_tr     <= L2r_tr and L2_tw  when trigger_input_mask = '0' else '0';
  L0_trigger        <= L0_masked_tr  when (disable_error_mask = '0') else L0;
  L1a_trigger       <= L1a_masked_tr when (disable_error_mask = '0') else L1a;
  L1r_trigger       <= L1r_tr            when trigger_input_mask = '0' else '0';
  L2a_trigger       <= L2a_masked_tr when (disable_error_mask = '0') else L2a_tr;
  L2r_trigger       <= L2r_masked_tr when (disable_error_mask = '0') else L2r_tr;
  L2_timeout        <= L2_timeout_i      when trigger_input_mask = '0' else '0';
  L2_trigger        <= (L2a_masked_tr and L2a_mask) or (L2r_masked_tr and L2r_mask) or (L2_timeout_i and L2_timeout_mask);
  DAQ_header_out    <= DAQ_headerFIFO;
  
  -- trigger that is checked with phase check tool. 
  trigger           <= L0_masked_tr when detector = '1' else L1a_masked_tr;
  -- soft reset
  soft_reset_i      <= module_reset_i or soft_reset;
  
  -- combining for use in setting time window counter
  receiving_L2      <= receiving_l2a or receiving_l2r;
  
  -- multi event buffer mask. 
  p_trigger_input_mask : process (clk, reset_n)
  begin
    if (reset_n = '0') then
      trigger_input_mask <= '0';
    elsif rising_edge(clk) then
      if (trigger_input_mask_enable = '1') then
        -- set when max number of multi event buffers are used 
        if (buffered_events_i = X"8" or ddl_rdy_rx = '0' or meb_full = '1') then
          trigger_input_mask <= '1';
        -- released when not busy receiving the ongoing sequence. This is done to make sure the complete sequence is masked regardless of when the FIFO is being read...
        elsif (busy_i = '0') then 
          trigger_input_mask <= '0';
        end if;
      else
        trigger_input_mask <= '0';
      end if;
    end if;
  end process p_trigger_input_mask;
  
  -- assigning internal signals
  OrbitID           <= L2a_message(1) & L2a_message(2);
  L2Class           <= L2a_message(3)(1 downto 0) & L2a_message(4) & L2a_message(5) & L2a_message(6) & L2a_message(7);
  BCID              <= BCID_L2r when (event_info_reg(10) = '1') else
                       BCID_L2a when (event_info_reg(11) = '1') else
                       (others => '0');
  header_version    <= c_headerversion;

  -- EVENT INFO REGISTER
  event_info_reg    <= X"A95" & "00" & phase & seq_info_reg; -- synch word - phase counter - info
  -- EVENT ERROR REGISTER
  event_error_reg   <= X"0" & "000" & seq_error_reg & msg_error_reg & hmg_error_reg & serialB_error_reg;
  
  ------------------------------------------------------------------------------------------------
  -- setting up DAQ error/status register
  ------------------------------------------------------------------------------------------------
  DAQ_errorStatus(0)  <= event_error_reg(10); -- L0 Trigger overlap error
  DAQ_errorStatus(1)  <= event_error_reg(11); -- Missing L0 trigger
  DAQ_errorStatus(2)  <= '0'; -- Data parity error - not handled here
  DAQ_errorStatus(3)  <= '0'; -- Control parity error - not handled here
  DAQ_errorStatus(4)  <= event_error_reg(6) or event_error_reg(7) or event_error_reg(8) or  -- Incomplete L1, L2 or RoI message
                         event_error_reg(16) or event_error_reg(18) or event_error_reg(20); -- Missing L1, L2 or RoI message
  DAQ_errorStatus(5)  <= '0'; -- FEE error - not handled here.
  DAQ_errorStatus(6)  <= '0'; -- HLT decision flag - not handled here.
  DAQ_errorStatus(7)  <= '0'; -- HLT payload flag  - not handled here.
  DAQ_errorStatus(8)  <= '0'; -- DDG payload flag  - not handled here.
  DAQ_errorStatus(9)  <= event_error_reg(12); -- spurious L1
  DAQ_errorStatus(10) <= (event_error_reg(17) or event_error_reg(18)) and (not event_error_reg(14));
                         -- L2 timeout or L2 not received completely within legal timeslot when L1 trigger has arrived.
  DAQ_errorStatus(11) <= event_error_reg(21); -- pre pulse error (always zero as this is not detectable)
  DAQ_errorStatus(12) <= event_error_reg(0) or event_error_reg(1) or event_error_reg(2) or event_error_reg(3) or
                         event_error_reg(4) or event_error_reg(5) or event_error_reg(13) or event_error_reg(15) or 
                         event_error_reg(19) or event_error_reg(22) or event_error_reg(23) or event_error_reg(24); -- all other errors ored.
  DAQ_errorStatus(13) <= event_error_reg(14) and (not event_error_reg(18)); -- L1 trigger not arrived when L2 arrived.
  DAQ_errorStatus(14) <= '0'; -- Multi-Event Buffer Error - not handled - not sure how to
  DAQ_errorStatus(15) <= '0'; -- Future use
  ------------------------------------------------------------------------------------------------

  ------------------------------------------------------------------------------------------------
  -- Setting up DAQ headers
  ------------------------------------------------------------------------------------------------
  DAQ_header(0)     <= X"0" & header_version & "000" & CIT_L1 & RoC & ESR_L1 & L1_SwC & "00" & BCID;
  DAQ_header(1)     <= X"00" & OrbitID;
  DAQ_header(2)     <= X"00" & L2Class(47 downto 24);
  DAQ_header(3)     <= X"0"  & DAQ_errorStatus & BCID_local;
  DAQ_header(4)     <= L2Class(31 downto 0);
  DAQ_header(5)     <= X"000" & "00" & L2Class(49 downto 32);
  DAQ_header(6)     <= (others => '0');
  ------------------------------------------------------------------------------------------------
  
  serialBchannel_i <= serialBchannel when input_enable = '1' else '1';
  L1Accept_i       <= L1Accept       when input_enable = '1' else '0';

  -- setting up reset
  module_reset     <= module_reset_i;
  read_enable_i    <= read_enable;
  -- purpose: Implements an interface towards the DCS module of the RCU FW.
  -- Includes control registers to set up the module correctly from the DCS board,
  -- and returns information back.
  RCU_bus_communication : entity work.rcu_com_release
  port map(
    clk                 => clk,
    reset_n             => reset_n,
    data_in             => data_in,
    data_out            => data_out,
    RCU_addr            => RCU_addr,
    we                  => we,
    input_enable        => input_enable,
    L2a_mask            => L2a_mask,
    L2r_mask            => L2r_mask,
    L2_timeout_mask     => L2_timeout_mask,
    counter_reset       => counter_reset,
    module_reset        => module_reset_i,
    disable_error_mask  => disable_error_mask,
    trigger_input_mask_enable     => trigger_input_mask_enable,
    L1a_msg_mask        => L1a_msg_mask,
    L0_support          => L0_support,
    header_version      => header_version,
    L1_Latency          => L1_Latency,
    L1_boundary         => L1_boundary,
    L2_Latency          => L2_Latency,
    L1_msg_Latency      => L1_msg_Latency,
    bc_overflow         => BunchCnt_overflow_i,
    busy                => busy_i,
    run_active          => run_active_i,
    BCID_Local          => BCID_local,
    pp_counter          => pp_counter,
    L0_Counter          => L0_Counter,
    L1_Counter          => L1_Counter,
    L2a_Counter         => L2a_Counter,
    L2r_Counter         => L2r_Counter,
    L1_msg_counter      => L1_msg_counter,
    hammingErrorCnt     => hammingError_counter,
    msgErrorCnt         => msgError_counter,
    seqErrorCnt         => seqError_counter,
    buffered_events     => buffered_events_i,
    event_info_reg      => event_info_reg(17 downto 0),
    event_error_reg     => event_error_reg(24 downto 0),
    DAQ_header          => DAQ_header);

    -- Purpose: Decodes messages received on the serial B line into broadcast or individually adressed messages
  serialb_decoder: entity work.serialb_com
  port map(
    clk                 => clk,
    reset_n             => reset_n,
    module_reset        => soft_reset_i,
    serialBchannel      => serialBchannel_i,
    init_sequence       => init_sequence,
    error_reg           => serialB_error_reg,
    brcdata_ready       => brcdata_ready,
    adrdata_ready       => adrdata_ready,
    received_data       => received_data);

  -- Purpose: Combinatorial module that decodes rceived hamming code on the fly for the messages 
  -- from the serial B decoder. The messages are split into broadcast and addressed messages
  -- The errors are registered.
  hamming_decoder: entity work.hamming_decoder
  port map(
    clk                 => clk,
    reset_n             => reset_n,
    module_reset        => soft_reset_i,
    received_data       => received_data,
    brcdata_ready_in    => brcdata_ready,
    adrdata_ready_in    => adrdata_ready,
    init_sequence       => init_sequence,
    single_bit_error    => single_bit_error, -- must be saved
    double_bit_error    => double_bit_error, -- must be saved
    error_reg           => hmg_error_reg,
    brcdata_ready_out   => hmg_brcdata_ready,
    adrdata_ready_out   => hmg_adrdata_ready,
    brcdata_out         => brcdata,
    adrdata_out         => adrdata);

  -- Purpose: Decodes the broadcast messages into:
  -- pre-pulse, bunch counter reset & event counter reset.
  -- If more broadcast messages are added later on - this is where they should be handled.
  broadcast_message_decoder: entity work.broadcast_message_decoder
  port map(
    clk                 => clk,
    reset_n             => reset_n,
    brcdata             => brcdata,
    brcdata_ready       => hmg_brcdata_ready,
    bunchCnt_reset      => bunchCnt_reset,
    eventCnt_reset      => eventCnt_reset,
    cal_pre_pulse       => pre_pulse);

  -- Purpose: Decodes the addressed messages into:
  -- L1a message, RoI message, L2a message / L2r message
  -- All other message adresses are unknown to the module.
  -- If more addressed messages are added later on - this is where they should be handled.
  addressed_message_decoder : entity work.addressed_message_decoder
  port map(
    clk                 => clk,
    reset_n             => reset_n,
    module_reset        => soft_reset_i,
    adrdata             => adrdata,
    adrdata_ready       => hmg_adrdata_ready,
    init_sequence       => init_sequence,
    L1a_msg_mask        => L1a_msg_mask,
    L2_trigger          => L2_trigger,
    error_reg           => msg_error_reg,
    msg_error           => msg_error,
    receiving_L1        => receiving_L1,
    receiving_L2a       => receiving_L2a,
    receiving_L2r       => receiving_L2r,
    FEE_reset           => FEE_reset,
    L1_message          => L1_message,
    L2a_message         => L2a_message,
    L2r_message         => L2r_message);

  -- Purpose: Decodes data on the L1Accpt line into L0 or L1 triggers
  L1_line_decoder : entity work.L1_line_decoder
  port map(
    clk                 => clk,
    reset_n             => reset_n,
    L0_support          => L0_support,
    L1Accept            => L1Accept_i,
    L0                  => L0,
    L1a                 => L1a);

  -- Purpose: Validates if the sequence is correctly received. This implies both that
  -- it starts with the correct trigger, that the timing in the reception of the messages are 
  -- correct and that information in the messages are correct.
  -- Issues the triggers based on the message decoding.
  Sequence_validator : entity work.sequence_validator
  port map(
    clk                 => clk,
    reset_n             => reset_n,
    module_reset        => soft_reset_i,
    pre_pulse           => pre_pulse,
    L0                  => L0,
    L1                  => L1a,
    receiving_L1        => receiving_L1,
    receiving_L2a       => receiving_L2a,
    receiving_L2r       => receiving_L2r,
    L1a_msg_mask        => L1a_msg_mask,
    BCID_L2a            => BCID_L2a,
    BCID_L2r            => BCID_L2r,
    ESR_L1              => ESR_L1,
    ESR_L2a             => ESR_L2a,
    L1_SwC              => L1_SwC,
    L2_SwC              => L2_SwC,
    CIT_L1              => CIT_L1,
    CIT_L2a             => CIT_L2a,
    RoC                 => RoC,
    L0_tw               => L0_tw,
    L1_tw               => L1_tw,
    L1_exact_tw         => L1_exact_tw,
    L1_msg_tw           => L1_msg_tw,
    L2_tw               => L2_tw,
    L1_tw_passed        => L1_tw_passed,
    L1_msg_tw_passed    => L1_msg_tw_passed,
    L2_tw_passed        => L2_tw_passed,
    info_reg            => seq_info_reg,
    error_reg           => seq_error_reg,
    init_sequence       => init_sequence,
    busy                => busy_i,
    sequence_error      => sequence_error,
    start_of_run        => start_of_run,
    end_of_run          => end_of_run,
    sync                => sync,
    run_active          => run_active_i,
    L1r_out             => L1r_tr,
    L1_msg_out          => L1_msg_tr,
    L2a_out             => L2a_tr,
    L2r_out             => L2r_tr,
    L2_timeout          => L2_timeout_i);

  --test what part of sampling clock phase the trigger is received
  phase_check: if include_phase_check = 1 generate
    phase_checker: entity work.phase_check
    port map(
      clk                 => clk,-- 40 MHz system clock
      sclk                => sclk,-- sampling clock (2,5 - 10 MHz)
      reset_n             => reset_n, -- asynch reset, negative ploarity
      trigger             => trigger, -- L0 or L1 trigger depending on PHOS or TPC mode
      clear               => soft_reset_i,-- phase is only cleared by reset
      phase               => phase); -- counter giving phase of when trigger is received.
  end generate phase_check;
  
  no_phase_check: if not (include_phase_check = 1) generate
    phase <= (others => '0');
  end generate no_phase_check;
  
  
  -- purpose: When an event is finished with or withour errors, this module stores it in the FIFO 
  -- and increments the buffered_event counter. The data parity bit is also set in this module.
  FIFO: entity work.fifo_wrapper
  generic map(
    include_parity      => include_fifo_parity,
	platform => platform
	)
  port map(
    clk                 => clk,
    reset_n             => reset_n,
    module_reset        => soft_reset_i,
    read_enable         => read_enable_i,
    data_out            => DAQ_headerFIFO,
    read_counter_out    => read_counter_out,
    buffered_events     => buffered_events_i,
    event_info_reg      => event_info_reg,
    event_error_reg     => event_error_reg,
    DAQ_header          => DAQ_header,
    L2_trigger          => L2_trigger);

  
  -- Purpose: Counts all triggers, messages and errors. Also generates time windows for 
  -- legal reception of messages and triggers. 
  -- Stores local BCID.
  Counters: entity work.counters
  port map(
    clk                 => clk,
    reset_n             => reset_n,
    module_reset        => soft_reset_i,
    pre_pulse           => pre_pulse,
    L0                  => L0,
    L1                  => L1a,
    L1_msg              => L1_msg_tr,
    L2A                 => L2a_tr,
    L2R                 => L2r_tr,
    init_sequence       => init_sequence,
    receiving_L1        => receiving_L1,
	  receiving_L2        => receiving_L2, 
	  busy                => busy_i,
    counter_reset       => counter_reset,
    reset_bunch         => bunchCnt_reset,
    single_bit_error    => single_bit_error,
    double_bit_error    => double_bit_error,
    msg_error           => msg_error,
    sequence_error      => sequence_error,
    L1_Latency          => L1_Latency,
    L1_boundary         => L1_boundary,
    L2_Latency          => L2_Latency,
    L1_msg_Latency      => L1_msg_Latency,
    L0_tw               => L0_tw,
    L1_tw               => L1_tw,
    L1_exact_tw         => L1_exact_tw,
    L1_msg_tw           => L1_msg_tw,
    L2_tw               => L2_tw,
    L1_tw_passed        => L1_tw_passed,
    L1_msg_tw_passed    => L1_msg_tw_passed,
    L2_tw_passed        => L2_tw_passed,
    BunchCnt_overflow   => BunchCnt_overflow_i,
    Bunch_Counter       => Bunch_Counter,
    BCID_Local          => BCID_Local,
    pp_Counter          => pp_counter,
    L0_Counter          => L0_counter,
    L1_Counter          => L1_counter,
    L1_msg_Counter      => L1_msg_counter,
    L2a_Counter         => L2a_counter,
    L2r_Counter         => L2r_counter,
    hammingErrorCnt     => hammingError_counter,
    msgErrorCnt         => msgError_counter,
    seqErrorCnt         => seqError_counter);

end arc;
